from .model import Model  # noqa: F401
from .muscle import Muscle  # noqa: F401
from .muscle import PotvinMuscle  # noqa: F401
from .potvin_2017_muscle_fibers import Potvin2017MuscleFibers  # noqa: F401
from .potvin_2017_motor_neuron_pool import Potvin2017MotorNeuronPool  # noqa: F401
from .__version__ import __version__  # noqa: F401



import numpy as np
import math # noqa
from numpy import ndarray
from copy import copy

from .model import Model


class Potvin2017MuscleFibers(Model):
    """
    Encapsulates the muscle fibers portions of the motor unit model.

    The name of each parameter as it appears in Potvin, 2017 is in parentheses.
    If a parameter does not appear in the paper but does appear in the Matlab
    code, the variable name from the Matlab code is in parentheses.

    :param motor_unit_count: Number of motor units in the muscle (n)
    :param max_twitch_amplitude: Max twitch force within the pool (RP)
    :param max_contraction_time:
        [milliseconds] Maximum contraction time for a motor unit (tL)
    :param contraction_time_range:
        The scale between the fastest contraction time and the slowest (rt)
    :fatigue_factor_first_unit:
        The nominal fatigability of the first motor unit in percent / second
    :fatigability_range:
        The scale between the fatigability of the first motor unit and the last
    :contraction_time_change_ratio:
        For each percent of force lost during fatigue, what percentage should
        contraction increase? Based on Shields et al (1997)

    .. todo::
        The argument naming isn't consistent. Sometimes we use 'max' and other
        times we use 'last unit'. Can these be made consistent?

    Usage::

      from pymuscle import Potvin2017MuscleFibers as Fibers

      motor_unit_count = 60
      fibers = Fibers(motor_unit_count)
      motor_neuron_firing_rates = np.rand(motor_unit_count) * 10.0
      force = fibers.step(motor_neuron_firing_rates)
    """
    def __init__(
        self,
        motor_unit_count: int = 120,
        max_twitch_amplitude: int = 100,
        max_contraction_time: int = 90,
        contraction_time_range: int = 3,
        max_recruitment_threshold: int = 50,
        fatigue_factor_first_unit: float = 0.0125,
        max_fatigue_rate: float = 0.0225,
        fatigability_range: int = 180,
        contraction_time_change_ratio: float = 0.379,
        apply_fatigue: bool = True
    ):
        self._peak_twitch_forces = self._calc_peak_twitch_forces(
            motor_unit_count,
            max_twitch_amplitude
        )

        # These will change with fatigue.
        self._current_twitch_forces = copy(self._peak_twitch_forces)

        self._contraction_times = self._calc_contraction_times(
            max_twitch_amplitude,
            max_contraction_time,
            contraction_time_range,
            self._peak_twitch_forces
        )

        # These will change with fatigue
        self._current_contraction_times = copy(self._contraction_times)

        # The maximum rates at which motor units will fatigue
        self._nominal_fatigabilities = self._calc_nominal_fatigabilities(
            motor_unit_count,
            fatigability_range,
            max_fatigue_rate,
            self._peak_twitch_forces
        )

        # Assing other non-public attributes
        self._contraction_time_change_ratio = contraction_time_change_ratio
        self._apply_fatigue = apply_fatigue

        # Assign public attributes
        self.motor_unit_count = motor_unit_count

    def _update_fatigue(
        self,
        normalized_forces: ndarray,
        step_size: float
    ) -> None:
        """
        Updates current twitch forces and contraction times.

        :param normalized_forces:
            Array of scaled forces. Used to weight how much fatigue will be
            generated in this step.
        :param step_size: How far time has advanced in this step.
        """
        # Instantaneous fatigue rate
        fatigues = (self._nominal_fatigabilities * normalized_forces) * step_size
        self._current_twitch_forces -= fatigues
        # Zero out negative values
        self._current_twitch_forces[self._current_twitch_forces < 0] = 0.0
        self._update_contraction_times()

    def _apply_recovery(self):
        """
        Updates twitch forces and contraction times.
        """
        raise NotImplementedError

    def _update_contraction_times(self) -> None:
        """
        Update our current contraction times as a function of our current
        force capacity relative to our peak force capacity.
        From Eq. (11)
        """
        force_loss_pcts = 1 - (self._current_twitch_forces / self._peak_twitch_forces)
        inc_pcts = 1 + self._contraction_time_change_ratio * force_loss_pcts
        self._current_contraction_times = self._contraction_times * inc_pcts

    def _calc_ct_increase_percents(self, force_losses: ndarray):
        """
        Calculates the percent that contraction times should increase for
        a given loss in force.

        :param force_losses: Percentage decrease in force being produced
            by each motor unit
        """
        return self._contraction_time_change_ratio * force_losses

    @staticmethod
    def _calc_contraction_times(
        max_twitch_amplitude: int,
        max_contraction_time: int,
        contraction_time_range: int,
        peak_twitch_forces: ndarray
    ) -> ndarray:
        """
        Calculate the contraction times for each motor unit
        Results in a smooth range from max_contraction_time at the first
        motor unit down to max_contraction_time / contraction_time range
        for the last motor unit

        :param max_twitch_amplitude:
            Largest force a motor unit in this muscle can produce.
        :param max_contraction_time:
            Slowest contraction time for a motor unit in this muscle.
        :param contraction_time_range:
            The ratio between the slowest and the fastest contraction times
            in this muscle.
        :param peak_twitch_forces:
            An array of all the largest forces that each motor unit can
            produce.
        """

        # Fuglevand 93 version - very slightly different values
        # twitch_force_range = peak_twitch_forces[-1] / peak_twitch_forces[0]
        # scale = math.log(twitch_force_range, contraction_time_range)

        # Potvin 2017 version
        scale = np.log(max_twitch_amplitude) / np.log(contraction_time_range)

        mantissa = 1 / peak_twitch_forces
        exponent = 1 / scale
        return max_contraction_time * np.power(mantissa, exponent)

    @staticmethod
    def _calc_peak_twitch_forces(
        motor_unit_count: int,
        max_twitch_amplitude: int
    ) -> ndarray:
        """
        Pure function to calculate the peak twitch force for each motor unit.

        :param motor_unit_count: The number of motor units in the pool.
        :param max_twitch_amplitude:
            Largest force a motor unit in this muscle can produce.
        """
        motor_unit_indices = np.arange(1, motor_unit_count + 1)
        t_log = np.log(max_twitch_amplitude)
        t_exponent = (t_log * (motor_unit_indices - 1)) / (motor_unit_count - 1)
        return np.exp(t_exponent)

    @staticmethod
    def _calc_nominal_fatigabilities(
        motor_unit_count: int,
        fatigability_range: int,
        max_fatigue_rate: float,
        peak_twitch_forces: ndarray
    ) -> ndarray:
        """
        Pure function to calculate nominal fatigue factors for each motor unit.

        Taken more from the matlab code than the paper.

        :param motor_unit_count: The number of motor units in this muscle.
        :param fatigability_range:
            The ratio between the maximum fatigue rate of the strongest motor
            unit (which fatigues the fastest) and the fatigue rate of the
            weakest motor unit.
        :param max_fatigue_rate:
            Largest percentage drop per unit time of twitch strength for the
            strongest motor unit.
        :param peak_twitch_forces:
            An array of all the largest forces that each motor unit can
            produce.
        """
        motor_unit_indices = np.arange(1, motor_unit_count + 1)
        f_log = np.log(fatigability_range)
        motor_unit_fatigue_curve = np.exp((f_log / (motor_unit_count - 1)) * (motor_unit_indices - 1))
        fatigue_rates = motor_unit_fatigue_curve * (max_fatigue_rate / fatigability_range) * peak_twitch_forces
        return fatigue_rates

    def _normalize_firing_rates(self, firing_rates: ndarray) -> ndarray:
        """
        Calculate the effective impact of a given set of firing rates on
        muscle fibers which have diverse contraction times and may be fatigued.

        :param firing_rates: Should be the result of pool._calc_adapted_firing_rates()
        """
        # Divide by 1000 here as firing rates are per second where contraction
        # times are in milliseconds.
        return self._current_contraction_times * (firing_rates / 1000)

    @staticmethod
    def _calc_normalized_forces(normalized_firing_rates: ndarray) -> ndarray:
        """
        Calculate motor unit force, relative to its peak force. Force grows
        in a linear fashion up to 0.4 normalized firing rate and then in a
        sigmoid curve afterward.

        :param normalized_firing_rates:
            An array of firing rates scaled by the current contraction times
            for each motor unit.
        """
        normalized_forces = copy(normalized_firing_rates)
        linear_threshold = 0.4  # Values are non-linear above this value
        below_thresh_indices = normalized_forces <= linear_threshold
        above_thresh_indices = normalized_forces > linear_threshold
        # The next two lines are strange and magical
        # In the paper they are simplified to *= 0.3
        # This is the equivalent of the Matlab code
        normalized_forces[below_thresh_indices] /= 0.4
        normalized_forces[below_thresh_indices] *= 1 - np.exp(-2 * (0.4 ** 3))
        exponent = -2 * np.power(
            normalized_forces[above_thresh_indices],
            3
        )
        normalized_forces[above_thresh_indices] = 1 - np.exp(exponent)
        return normalized_forces

    def _calc_current_forces(self, normalized_forces: ndarray) -> ndarray:
        """
        Scales the normalized forces for each motor unit by their current
        remaining twitch force capacity.

        :param normalized_forces: An array of forces scaled between 0 and 1
        """
        return normalized_forces * self._current_twitch_forces

    def _calc_total_fiber_force(
        self,
        firing_rates: ndarray,
        step_size: float
    ) -> ndarray:
        """
        Calculates the total instantaneous force produced by all fibers for
        the given instantaneous firing rates.
        :param firing_rates:
            An array of firing rates calculated by a compatible Pool class.
        :param step_size: How far time has advanced in this step.
        """
        normalized_firing_rates = self._normalize_firing_rates(firing_rates)
        normalized_forces = self._calc_normalized_forces(normalized_firing_rates)
        current_forces = self._calc_current_forces(normalized_forces)
        # Apply fatigue as last step
        if self._apply_fatigue:
            self._update_fatigue(normalized_forces, step_size)

        total_force = np.sum(current_forces)
        return total_force

    def step(
        self,
        motor_pool_output: ndarray,
        step_size: float = 0.1
    ) -> float:
        """
        Advance the muscle fibers simulation one step.

        Returns the total instantaneous force produced by all fibers for
        the given input from the motor neuron pool.

        :param motor_pool_output:
            An array of firing rates calculated by a compatible Pool class.
        :param step_size: How far time has advanced in this step.
        """
        return self._calc_total_fiber_force(motor_pool_output, step_size)



import numpy as np
from numpy import ndarray

from .model import Model


class Potvin2017MotorNeuronPool(Model):
    """
    Encapsulates the motor neuron portion of the motor unit model.

    The name of each parameter as it appears in Potvin, 2017 is in parentheses.
    If a parameter does not appear in the paper but does appear in the
    accompanying Matlab code, the variable name from the Matlab code is used in
    the parentheses.

    :param motor_unit_count: Number of motor units in the muscle (n)
    :param max_recruitment_threshold:
        Max excitation required by a motor unit within the pool before
        firing (RR)
    :param firing_gain:
        The slope of firing rate by excitation above threshold (g)
    :param min_firing_rate:
        The minimum firing rate for a motor neuron above threshold (minR)
    :param max_firing_rate_first_unit:
        Max firing rate for the first motor unit (maxR(1))
    :param max_firing_rate_last_unit:
        Max firing rate for the last motor unit (maxR(last))
    :param pre_calc_firing_rates:
        Whether to build a dict mapping excitation levels to firing rates for
        each motor neuron. This can speed up simulation at the cost of
        additional memory.
    :param pre_calc_resolution:
        Step size for excitation levels to pre-calculate (res)
    :param pre_calc_max: Highest excitation value to pre-calculate
    :param derecruitment_delta:
        Absolute minimum firing rate = min_firing_rate - derecruitment_delta
        (d)
    :param adaptation_magnitude:
        Magnitude of adaptation for different levels of excitation.(phi)
    :param adaptation_time_constant:
        Time constant for motor neuron adaptation (tau). Default based on
        Revill & Fuglevand (2011)
    :param max_duration:
        Longest duration of motor unit activity that will be recorded. The
        default value should be >> than the time it takes to fatigue all
        fibers. Helps prevent unbounded values.

    .. todo::
        Make pre_calc_max a function of other values as in the matlab code.
        This will also require changing how we look up values if they
        are larger than this value.

    Usage::

      from pymuscle import Potvin2017MotorNeuronPool as Pool

      motor_unit_count = 60
      pool = Pool(motor_unit_count)
      excitation = np.full(motor_unit_count, 10.0)
      step_size = 1 / 50.0
      firing_rates = pool.step(excitation, step_size)
    """
    def __init__(
        self,
        motor_unit_count: int = 120,
        max_recruitment_threshold: int = 50,
        firing_gain: float = 1.0,
        min_firing_rate: int = 8,
        max_firing_rate_first_unit: int = 35,
        max_firing_rate_last_unit: int = 25,
        pre_calc_firing_rates: bool = False,
        pre_calc_resolution: float = 0.1,
        pre_calc_max: float = 70.0,
        derecruitment_delta: int = 2,
        adaptation_magnitude: float = 0.67,
        adaptation_time_constant: float = 22.0,
        max_duration: float = 20000.0,
        apply_fatigue: bool = True
    ):
        self._recruitment_thresholds = self._calc_recruitment_thresholds(
            motor_unit_count,
            max_recruitment_threshold
        )

        self._peak_firing_rates = self._calc_peak_firing_rates(
            max_firing_rate_first_unit,
            max_firing_rate_last_unit,
            max_recruitment_threshold,
            self._recruitment_thresholds
        )

        self._recruitment_durations = np.zeros(motor_unit_count)

        # Assign additional non-public attributes
        self._max_recruitment_threshold = max_recruitment_threshold
        self._firing_gain = firing_gain
        self._min_firing_rate = min_firing_rate
        self._derecruitment_delta = derecruitment_delta
        self._adaptation_magnitude = adaptation_magnitude
        self._adaptation_time_constant = adaptation_time_constant
        self._max_duration = max_duration
        self._apply_fatigue = apply_fatigue

        # Assign public attributes
        self.motor_unit_count = motor_unit_count

        # Pre-calculate firing rates for all motor neurons across a range of
        # possible excitation levels.
        self._firing_rates_by_excitation = None
        if pre_calc_firing_rates:
            self._build_firing_rates_cache(pre_calc_max, pre_calc_resolution)

    def _build_firing_rates_cache(
        self,
        pre_calc_max: float,
        pre_calc_resolution: float
    ) -> None:
        """
        Pre-calculate and store firing rates for all motor neurons across a
        range of possible excitation levels.

        :param pre_calc_max: The maximum excitation value to calculate.
        :param pre_calc_resolution: 
            The step size between values during calculations.
        """

        self._firing_rates_by_excitation = {}
        # TODO: This is a hack. Maybe memoize vs pre-calculate?
        # Maybe https://docs.python.org/3/library/functools.html#functools.lru_cache
        resolution_places = len(str(pre_calc_resolution).split(".")[1])
        excitations = np.zeros(self.motor_unit_count)
        excitation_values = np.arange(
            0.0,
            pre_calc_max + pre_calc_resolution,
            pre_calc_resolution
        )
        for i in excitation_values:
            i = round(i, resolution_places)
            excitations += pre_calc_resolution
            self._firing_rates_by_excitation[i] = \
                self._inner_calc_firing_rates(
                    excitations,
                    self._recruitment_thresholds,
                    self._firing_gain,
                    self._min_firing_rate,
                    self._peak_firing_rates
            )

    def _calc_adapted_firing_rates(
        self,
        excitations: ndarray,
        step_size: float
    ) -> ndarray:
        """
        Calculate the firing rate for the given excitation including motor
        neuron fatigue (adaptation).

        :param excitations:
            Array of excitation levels to use as input to motor neurons.
        :param step_size: How far to advance time in this step.

        """
        firing_rates = self._calc_firing_rates(excitations)
        self._update_recruitment_durations(firing_rates, step_size)
        adaptations = self._calc_adaptations(firing_rates)

        adapted_firing_rates = firing_rates - adaptations
        return adapted_firing_rates

    def _calc_firing_rates(self, excitations: ndarray) -> ndarray:
        """
        Calculates firing rates on a per motor neuron basis for the given
        array of excitations.

        :param excitations:
            Array of excitation levels to use as input to motor neurons.
        """
        assert (len(excitations) == len(self._recruitment_thresholds))

        if self._firing_rates_by_excitation:
            excitation = excitations[0]  # TODO - Support variations
            firing_rates = self._firing_rates_by_excitation[excitation]
        else:
            firing_rates = self._inner_calc_firing_rates(
                excitations,
                self._recruitment_thresholds,
                self._firing_gain,
                self._min_firing_rate,
                self._peak_firing_rates
            )

        return firing_rates

    @staticmethod
    def _inner_calc_firing_rates(
        excitations: ndarray,
        thresholds: ndarray,
        gain: float,
        min_firing_rate: int,
        peak_firing_rates: ndarray
    ) -> ndarray:
        """
        Pure function to do actual calculation of firing rates.

        :param excitations:
            Array of excitation levels to use as input to motor neurons.
        :param thresholds:
            Array of minimum firing rates required for motor neuron activity
        :param gain: How firing rates scale with excitations
        :param peak_firing_rates: Maximum allowed firing rates per neuron
        """

        firing_rates = excitations - thresholds
        firing_rates += min_firing_rate
        below_thresh_indices = firing_rates < min_firing_rate
        firing_rates[below_thresh_indices] = 0
        firing_rates *= gain

        # Check for max values
        above_peak_indices = firing_rates > peak_firing_rates
        firing_rates[above_peak_indices] = peak_firing_rates[above_peak_indices]

        return firing_rates

    def _update_recruitment_durations(
        self,
        firing_rates: ndarray,
        step_size: float
    ) -> None:
        """
        Increment the on duration for each on motor unit by step_size

        :param firing_rates: Array of activities for each motor neuron.
        :param step_size: How far to advance time in this step.
        """
        # If we don't update durations, no fatigue will occur.
        if not self._apply_fatigue:
            return

        on = firing_rates > 0
        self._recruitment_durations[on] += step_size
        # TODO: Enable as a recovery mechanism
        # off = firing_rates = 0
        # self._recruitment_durations[off] -= 0
        # Prevent overflows
        over = self._recruitment_durations > self._max_duration
        self._recruitment_durations[over] = self._max_duration
        # Can't be less than zero
        under = self._recruitment_durations < 0
        self._recruitment_durations[under] = 0

    def _calc_adaptations(self, firing_rates: ndarray) -> ndarray:
        """
        Calculate the adaptation rates for each neuron based on current
        activity levels.

        :param firing_rates: Array of activities for each motor neuron.
        """
        adapt_curve = self._calc_adaptations_curve(firing_rates)
        # From Eq. (12)
        exponent = -1 * (self._recruitment_durations / self._adaptation_time_constant)
        adapt_scale = 1 - np.exp(exponent)
        adaptations = adapt_curve * adapt_scale
        # Zero out negative values
        adaptations[adaptations < 0] = 0.0
        return adaptations

    def _calc_adaptations_curve(self, firing_rates: ndarray) -> ndarray:
        """
        Calculates q(i) from Eq. (13). This is the baseline adaptation curve
        for each neuron based on activity.

        :param firing_rates: Array of activities for each motor neuron.
        """
        ratios = (self._recruitment_thresholds - 1) / (self._max_recruitment_threshold - 1)
        adaptations = self._adaptation_magnitude * (firing_rates - self._min_firing_rate + self._derecruitment_delta) * ratios
        return adaptations

    @staticmethod
    def _calc_peak_firing_rates(
        max_firing_rate_first_unit: int,
        max_firing_rate_last_unit: int,
        max_recruitment_threshold: int,
        recruitment_thresholds: ndarray,
    ) -> ndarray:
        """
        Calculate peak firing rates for each motor neuron

        :param max_firing_rate_first_unit:
            Maximum allowed activity of the 'first' neuron in the pool.
        :param max_firing_rate_last_unit:
            Maximum allowed activity of the 'last' neuron in the pool.
        :param max_recruitment_threshold:
            Excitation level above which the 'last' neuron in the pool will
            become active.
        :param recruitment_thresholds:
            Array of all excitation levels above which each neuron will become
            active.
        """
        firing_rate_range = max_firing_rate_first_unit - max_firing_rate_last_unit
        rates = max_firing_rate_first_unit \
            - (firing_rate_range
                * ((recruitment_thresholds - recruitment_thresholds[0])
                    / (max_recruitment_threshold - recruitment_thresholds[0])))
        return rates

    @staticmethod
    def _calc_recruitment_thresholds(
        motor_unit_count: int,
        max_recruitment_threshold: int
    ) -> ndarray:
        """
        Pure function to calculate recruitment thresholds for each motor
        neuron.

        :param motor_unit_count: The number of motor units in the pool.
        :param max_recruitment_threshold:
            Excitation level above which the 'last' neuron in the pool will
            become active.
        """
        motor_unit_indices = np.arange(1, motor_unit_count + 1)

        r_log = np.log(max_recruitment_threshold)
        r_exponent = (r_log * (motor_unit_indices - 1)) / (motor_unit_count - 1)
        return np.exp(r_exponent)

    def step(self, motor_pool_input: ndarray, step_size: float) -> ndarray:
        """
        Advance the motor neuron pool simulation one step.

        Returns firing rates on a per motor neuron basis for the given
        array of excitations. Takes into account fatigue over time.

        :param excitations:
            Array of excitation levels to use as input to motor neurons.
        :param step_size: How far to advance time in this step.
        """
        return self._calc_adapted_firing_rates(motor_pool_input, step_size)



import numpy as np
from typing import Union

from .potvin_2017_muscle_fibers import Potvin2017MuscleFibers as Fibers
from .potvin_2017_motor_neuron_pool import Potvin2017MotorNeuronPool as Pool
from .model import Model


class Muscle(object):
    """
    A user-created :class:`Muscle <Muscle>` object.

    Used to simulate the input-output relationship between motor neuron
    excitation and muscle fibers contractile state over time.

    :param motor_neuron_pool_model:
        The motor neuron pool implementation to use with this muscle.
    :param muscle_fibers_model:
        The muscle fibers model implementation to use with this muscle.
    :param motor_unit_count: How many motor units comprise this muscle.

    Usage::

        from pymuscle import (Muscle,
                              Potvin2017MotorNeuronPool as Pool,
                              Potvin2017MuscleFibers as Fibers)

        motor_unit_count = 60
        muscle = Muscle(
            Pool(motor_unit_count),
            Fibers(motor_unit_count),
        )
        excitation = 32.0
        force = muscle.step(excitation, 1 / 50.0)
    """

    def __init__(
        self,
        motor_neuron_pool_model: Model,
        muscle_fibers_model: Model,
    ):
        assert motor_neuron_pool_model.motor_unit_count == \
            muscle_fibers_model.motor_unit_count

        self._pool = motor_neuron_pool_model
        self._fibers = muscle_fibers_model

    @property
    def motor_unit_count(self):
        return self._pool.motor_unit_count

    def step(
        self,
        motor_pool_input: Union[float, np.ndarray],
        step_size: float
    ) -> float:
        """
        Advances the muscle model one step.

        :param motor_pool_input:
            Either a single value or an array of values representing the
            excitatory input to the motor neuron pool for this muscle
        :param step_size:
            How far to advance the simulation in time for this step.
        """

        # Expand a single input to the muscle to a full array
        if isinstance(motor_pool_input, float):
            motor_pool_input = np.full(
                self._pool.motor_unit_count,
                motor_pool_input
            )

        motor_pool_output = self._pool.step(motor_pool_input, step_size)
        return self._fibers.step(motor_pool_output)


class PotvinMuscle(Muscle):
    """
    A thin wrapper around :class:`Muscle <Muscle>` which pre-selects the
    Potvin fiber and motor neuron models.
    """

    def __init__(
        self,
        motor_unit_count: int,
        apply_fatigue: bool = True,
        pre_calc_firing_rates: bool = False
    ):
        pool = Pool(
            motor_unit_count,
            apply_fatigue=apply_fatigue,
            pre_calc_firing_rates=pre_calc_firing_rates
        )
        fibers = Fibers(
            motor_unit_count,
            apply_fatigue=apply_fatigue
        )

        super().__init__(
            motor_neuron_pool_model=pool,
            muscle_fibers_model=fibers
        )
